@@ -49,17 +49,17 @@ tp_motion_history_offset(struct tp_touch *t, int offset)
 
 struct normalized_coords
 tp_filter_motion(struct tp_dispatch *tp,
-		 const struct normalized_coords *unaccelerated,
+		 const struct device_float_coords *unaccelerated,
 		 uint64_t time)
 {
 	struct device_float_coords raw;
+	const struct normalized_coords zero = { 0.0, 0.0 };
 
-	if (normalized_is_zero(*unaccelerated))
-		return *unaccelerated;
+	if (device_float_is_zero(*unaccelerated))
+		return zero;
 
-	/* Temporary solution only: convert back to raw coordinates, but
-	 * make sure we're on the same resolution for both axes */
-	raw = tp_unnormalize_for_xaxis(tp, *unaccelerated);
+	/* Convert to device units with x/y in the same resolution */
+	raw = tp_scale_to_xaxis(tp, *unaccelerated);
 
 	return filter_dispatch(tp->device->pointer.filter,
 			       &raw, tp, time);
@@ -67,17 +67,17 @@ tp_filter_motion(struct tp_dispatch *tp,
 
 struct normalized_coords
 tp_filter_motion_unaccelerated(struct tp_dispatch *tp,
-			       const struct normalized_coords *unaccelerated,
+			       const struct device_float_coords *unaccelerated,
 			       uint64_t time)
 {
 	struct device_float_coords raw;
+	const struct normalized_coords zero = { 0.0, 0.0 };
 
-	if (normalized_is_zero(*unaccelerated))
-		return *unaccelerated;
+	if (device_float_is_zero(*unaccelerated))
+		return zero;
 
-	/* Temporary solution only: convert back to raw coordinates, but
-	 * make sure we're on the same resolution for both axes */
-	raw = tp_unnormalize_for_xaxis(tp, *unaccelerated);
+	/* Convert to device units with x/y in the same resolution */
+	raw = tp_scale_to_xaxis(tp, *unaccelerated);
 
 	return filter_dispatch_constant(tp->device->pointer.filter,
 					&raw, tp, time);
@@ -131,6 +131,22 @@ tp_motion_history_push(struct tp_touch *t)
 	t->history.index = motion_index;
 }
 
+static inline void
+tp_maybe_disable_hysteresis(struct tp_dispatch *tp, uint64_t time)
+{
+	/* If the finger is down for 80ms without seeing motion events,
+	   the firmware filters and we don't need a software hysteresis */
+	if (tp->nfingers_down >= 1 &&
+	    time - tp->hysteresis.last_motion_time > ms2us(80)) {
+		tp->hysteresis.enabled = false;
+		evdev_log_debug(tp->device, "hysteresis disabled\n");
+		return;
+	}
+
+	if (tp->queued & TOUCHPAD_EVENT_MOTION)
+		tp->hysteresis.last_motion_time = time;
+}
+
 static inline void
 tp_motion_hysteresis(struct tp_dispatch *tp,
 		     struct tp_touch *t)
@@ -138,15 +154,18 @@ tp_motion_hysteresis(struct tp_dispatch *tp,
 	int x = t->point.x,
 	    y = t->point.y;
 
+	if (!tp->hysteresis.enabled)
+		return;
+
 	if (t->history.count == 0) {
 		t->hysteresis_center = t->point;
 	} else {
 		x = evdev_hysteresis(x,
 				     t->hysteresis_center.x,
-				     tp->hysteresis_margin.x);
+				     tp->hysteresis.margin.x);
 		y = evdev_hysteresis(y,
 				     t->hysteresis_center.y,
-				     tp->hysteresis_margin.y);
+				     tp->hysteresis.margin.y);
 		t->hysteresis_center.x = x;
 		t->hysteresis_center.y = y;
 		t->point.x = x;
@@ -272,7 +291,9 @@ tp_begin_touch(struct tp_dispatch *tp, struct tp_touch *t, uint64_t time)
 	t->thumb.state = THUMB_STATE_MAYBE;
 	t->thumb.first_touch_time = time;
 	t->tap.is_thumb = false;
+	t->tap.is_palm = false;
 	assert(tp->nfingers_down >= 1);
+	tp->hysteresis.last_motion_time = time;
 }
 
 /**
@@ -323,11 +344,11 @@ tp_stop_actions(struct tp_dispatch *tp, uint64_t time)
 	tp_tap_suspend(tp, time);
 }
 
-struct normalized_coords
+struct device_coords
 tp_get_delta(struct tp_touch *t)
 {
-	struct device_float_coords delta;
-	const struct normalized_coords zero = { 0.0, 0.0 };
+	struct device_coords delta;
+	const struct device_coords zero = { 0.0, 0.0 };
 
 	if (t->history.count <= 1)
 		return zero;
@@ -337,7 +358,7 @@ tp_get_delta(struct tp_touch *t)
 	delta.y = tp_motion_history_offset(t, 0)->point.y -
 		  tp_motion_history_offset(t, 1)->point.y;
 
-	return tp_normalize_delta(t->tp, delta);
+	return delta;
 }
 
 static void
@@ -1264,14 +1285,6 @@ tp_need_motion_history_reset(struct tp_dispatch *tp)
 	if (tp->nfingers_down != tp->old_nfingers_down)
 		return true;
 
-	/* if we're transitioning between slots and fake touches in either
-	 * direction, we may get a coordinate jump
-	 */
-	if (tp->nfingers_down != tp->old_nfingers_down &&
-		 (tp->nfingers_down > tp->num_slots ||
-		 tp->old_nfingers_down > tp->num_slots))
-		return true;
-
 	/* Quirk: if we had multiple events without x/y axis
 	   information, the next x/y event is going to be a jump. So we
 	   reset that touch to non-dirty effectively swallowing that event
@@ -1531,6 +1544,9 @@ static void
 tp_handle_state(struct tp_dispatch *tp,
 		uint64_t time)
 {
+	if (tp->hysteresis.enabled)
+		tp_maybe_disable_hysteresis(tp, time);
+
 	tp_process_state(tp, time);
 	tp_post_events(tp, time);
 	tp_post_process_state(tp, time);
@@ -2057,7 +2073,7 @@ tp_switch_event(uint64_t time, struct libinput_event *event, void *data)
 		break;
 	case LIBINPUT_SWITCH_STATE_ON:
 		tp_suspend(tp, tp->device);
-		evdev_log_debug(tp->device, "%s: suspend touchpad\n", which);
+		evdev_log_debug(tp->device, "%s: suspending touchpad\n", which);
 		break;
 	}
 }
@@ -2432,6 +2448,7 @@ tp_init_accel(struct tp_dispatch *tp)
 	 */
 	tp->accel.x_scale_coeff = (DEFAULT_MOUSE_DPI/25.4) / res_x;
 	tp->accel.y_scale_coeff = (DEFAULT_MOUSE_DPI/25.4) / res_y;
+	tp->accel.xy_scale_coeff = 1.0 * res_x/res_y;
 
 	if (tp->device->model_flags & EVDEV_MODEL_LENOVO_X230 ||
 	    tp->device->model_flags & EVDEV_MODEL_LENOVO_X220_TOUCHPAD_FW81)
@@ -2917,8 +2934,9 @@ tp_init_hysteresis(struct tp_dispatch *tp)
 
 	res_x = tp->device->abs.absinfo_x->resolution;
 	res_y = tp->device->abs.absinfo_y->resolution;
-	tp->hysteresis_margin.x = res_x/2;
-	tp->hysteresis_margin.y = res_y/2;
+	tp->hysteresis.margin.x = res_x/2;
+	tp->hysteresis.margin.y = res_y/2;
+	tp->hysteresis.enabled = true;
 }
 
 static void
